// AR.C
// version 1.5b
/*
** file archive utility loosely modeled after Kernigan & Plauger
*/

#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include "ar.h"


FN    *fnhead = (FN *)NULL;
direct char  *hid = HID,
             *suf = SUF;
direct char  *mod,                    /* pointer to module name */
             *archfile;               /* name of archive file */
direct int   all = FALSE,             /* true to access old versions */
             supflag = FALSE,         /* true to supress compression */
             oldstyle = FALSE,        /* true to fallback to 11 bit compression */
             zflag = FALSE;           /* true if names come from stdin */

char *attrs[8]={"---","--r","-w-","-wr","e--","e-r","ew-","ewr"};
char *title="Ar V1.5b - Archive file manager\n";

char  *emalloc();

barline()
{
    static char *bar="=======================================";
    printf("%s%s\n", bar, bar);
}
/*page*/
/*
** process command modifiers
*/

proc_opt(p)
char   *p;
   {
   while (*p)
      switch (tolower(*p++))
         {
         case 'a' :                          /* all - even old ones */
            all = TRUE;
            break;

         case 'o' :                          /* oldstyle compression */
            oldstyle= TRUE;
            break;

         case 's' :                          /* supress compression */
            supflag = TRUE;
            break;

         case 'z' :                         /* get names from stdin */
            zflag = TRUE;
            break;

         default :
            help();
         }
   }


/*
** process the command
*/

proc_cmd(command, afp)
char   command;
FILE  *afp;
   {
   switch (command)
      {
      case 'd' :
         break;

      case 't' :                               /* table of contents */
         table(afp);
         break;

      case 'u' :                                   /* update or add */
         update(afp);
         break;

      case 'p' :                                 /* print member(s) */
         extract(afp, 0);
         break;

      case 'x' :                               /* extract member(s) */
         extract(afp, 1);
         break;

      default  :
         help();
      }
   }
/*page*/
/*
** extract file(s) from the archive
**  copy a file from archive and restore it's origional attrs
*/

extract(afp, flag)
FILE  *afp;
int   flag;                     /* 0 = listing, 1 = writing to file */
   {
   FILE     *ofp;                            /* assume just listing */
   HEADER   header;
   FN       *fnp;
   FILE     *spl_open();
   int      bitsz;
   long     c4tol();

   if(flag){
      show_head();
      printf("%-31sVer Bits   Date   Time   Attr%9s\n",
             "Name", "Size");
      barline();
   }
   if (fnhead == (FN *) NULL)
      stash_name("*");                     /* fake for special case */

   while ((get_header(afp, &header)) != EOF)
      {
      for (fnp = fnhead; fnp; fnp = fnp->fn_link)
         if ((patmatch(fnp->fn_name, header.a_name, TRUE) == TRUE &&
            header.a_stat == 0)
            || (header.a_stat != 0 && all == TRUE))
            break;
      if (fnp == 0)
         fseek(afp, header.a_size, 1);            /* file not found */
      else
         {
         if (!flag)
            copy_from(afp, stdout, &header);
         else
            {
            if(header.a_type==PLAIN)
                bitsz=0;
            else
                bitsz=11;
            if(header.a_type==COMP2)
                bitsz=12;
         printf("%-32s%2d%4d  %02d/%02d/%02d %02d:%02d %s%s %7ld",
                header.a_name, header.a_stat, bitsz,

                header.a_attr.fd_date[0],
                header.a_attr.fd_date[1], header.a_attr.fd_date[2],
                header.a_attr.fd_date[3], header.a_attr.fd_date[4],

                attrs[(header.a_attr.fd_attr>>3) & 7],
                attrs[header.a_attr.fd_attr & 7],

                c4tol(header.a_attr.fd_fsize));

            fflush(stdout);
            ofp = spl_open(&header);
            copy_from(afp, ofp, &header);
            set_fstat(ofp->_fd, &header.a_attr);
            fclose(ofp);
            printf("%10s", "Done\n");
            }
         }
      }
   }
/*page*/
/*
** list a table of contents for the archive file
**  only show files matching the search mask which are current
**  unless the all flag is set, whereupon we will show old ones too
*/

table(fp)
FILE  *fp;
   {
   HEADER   header;
   FN       *fnp;
   long     c4tol();
   int      bitsz, save1, save2;

   if (fnhead == (FN *) NULL)
      stash_name("*");                     /* fake for special case */

   show_head();
   printf("%-31sVer Bits%7s%8s  SizeOrg SizeNow Saving\n",
           "Name", "Date", "Attr");
   barline();
   while ((get_header(fp, &header)) != EOF)
      for (fnp = fnhead; fnp; fnp = fnp->fn_link)
         {
         if (patmatch(fnp->fn_name, header.a_name, TRUE) == TRUE
             && (header.a_stat == 0 || all == TRUE))
             {
             if(header.a_type==PLAIN)
                bitsz=0;
             else
                bitsz=11;
             if(header.a_type==COMP2)
                bitsz=12;

             save1 = ((1000 - ((1000 * header.a_size)/
                     (c4tol(header.a_attr.fd_fsize) + 1))) / 10) % 100;
             save2 = ((10000 - ((10000 * header.a_size)/
                     (c4tol(header.a_attr.fd_fsize) + 1))) / 10) % 10;

             printf("%-32s%2d%4d  %02d/%02d/%02d %s%s %7ld %7ld%4d.%1d%%\n",
                  header.a_name, header.a_stat, bitsz,

                  header.a_attr.fd_date[0],
                  header.a_attr.fd_date[1],
                  header.a_attr.fd_date[2],

                  attrs[(header.a_attr.fd_attr>>3) & 7],
                  attrs[header.a_attr.fd_attr & 7],

                  c4tol(header.a_attr.fd_fsize),
                  header.a_size,
                  save1, save2);
             }
         fseek(fp, header.a_size, 1);
         }
   }
/*page*/
/*
** add new files or replace existing files
*/

update(afp)
FILE  *afp;
   {
   FILE     *ifp;
   HEADER   header;
   FN       *fnp;
   int      synch;
   long     bytes, head_pos, tail_pos, copy_to(), c4tol();
   int      save1, save2;

   show_head();
   printf("%-33sDate    Attr  SizeOrg     SizeNow%12s\n",
          "Name", "Saving");
   barline();
   while ((get_header(afp, &header)) != EOF)
      {
      for (fnp = fnhead; fnp; fnp = fnp->fn_link)
         if (patmatch(fnp->fn_name, header.a_name, TRUE) == TRUE)
            {
            ++header.a_stat;                       /* mark it older */
            fseek(afp, (long) -sizeof(HEADER), 1);
            if ((fwrite(&header, sizeof(HEADER), 1, afp)) == NULL)
               fatal(errno, "write failure on delete\n");
            }
      fseek(afp, header.a_size, 1);
      }

   for (fnp = fnhead; fnp; fnp = fnp->fn_link)
      {
      if ((ifp = fopen(fnp->fn_name, "r")) == NULL)
            if (errno == 214)
                continue;                    /* a directory, we presume */
            else
            fatal(errno, "can't find %s\n", fnp->fn_name);
      printf("%-31s", fnp->fn_name);
      fflush(stdout);
      if (supflag)                 /* || ((synch = getw(ifp)) == MSYNC)) */
         header.a_type = PLAIN;
      else {
         if(oldstyle)
            header.a_type = COMP1;
         else
            header.a_type = COMP2;
      }
      strcpy(header.a_hid, hid);
      setmem(header.a_name, FNSIZ + 1, ' ');
      strcpy(header.a_name, fnp->fn_name);
      get_fstat(ifp->_fd, &header.a_attr);
      printf("%02d/%02d/%02d %s%s%8ld",header.a_attr.fd_date[0],
             header.a_attr.fd_date[1],
             header.a_attr.fd_date[2],

             attrs[(header.a_attr.fd_attr>>3) & 7],
             attrs[header.a_attr.fd_attr & 7],

             c4tol(header.a_attr.fd_fsize));

      fflush(stdout);
      header.a_stat = '\0';
      rewind(ifp);
      head_pos = ftell(afp);                     /* save for update */
redo:
      fwrite(&header, sizeof(HEADER), 1, afp);        /* skip ahead */
      bytes = head_pos + c4tol(header.a_attr.fd_fsize) + sizeof(HEADER);
      set_fsize(afp->_fd, bytes);                /* set to maximum size */
      header.a_size = copy_to(afp, ifp, &header);
      tail_pos = ftell(afp);
      if((header.a_size>(c4tol(header.a_attr.fd_fsize))))
      {
         fseek(afp,head_pos,0);  /* back up to header */
         rewind(ifp);            /* back up to beggining */
         header.a_type=PLAIN;    /* switch to plain */
         goto redo;              /* go redo it */
      }
      fclose(ifp);
      save1 = ((1000 - ((1000 * header.a_size)/
              (c4tol(header.a_attr.fd_fsize) + 1))) / 10) % 100;
      save2 = ((10000 - ((10000 * header.a_size)/
              (c4tol(header.a_attr.fd_fsize) + 1))) / 10) % 10;
      printf("%12ld%9d.%1d%%\n",header.a_size,save1, save2);
      fflush(stdout);
      fseek(afp, head_pos, 0);             /* back up to header pos */
      if ((fwrite(&header, sizeof(HEADER), 1, afp)) == NULL)
         fatal(errno, "write error on header for %s\n", fnp->fn_name);
      fseek(afp, tail_pos, 0);                 /* go to end of file */
      }
   set_fsize(afp->_fd, tail_pos);         /* now set real file size */
   }
/*page*/
/*
** gather file names from command line or std in
**  use linked list to avoid finite limit on number of names
*/

get_names(ac, av, updating)
int   ac;
char  **av;
int   updating;                        /* TRUE if command is update */
   {
   char     *p, *q, *r, buf[80], *strhcpy();
   int      dirfd, found = 0;
   DIRENT   *dirp, *nextdir();
   char *rindex();
   char *index();
   
   while (ac--)
      if (!updating || !(index(*av, '?') || index(*av, '*')))
         found += stash_name(*av++);
      else
         {
         *(p = buf) = '\0';
         if (q = rindex((r = *av++), '/'))
            {
            strcpy(p, r);                     /* copy all to buffer */
            *(r = rindex(p, '/')) = '\0';           /* break in two */
            ++q;                              /* pointer to pattern */
            }
         else
            {
            q = r;
            r = p;                                 /* swap pointers */
            }
         dirfd = opndir(*p ? p : ".");
         if (*p)
            *r++ = '/';                        /* set up for append */
         while ((int) (dirp = nextdir(dirfd)) != -1)
            if (patmatch(q, strhcpy(r, dirp->dir_name), TRUE))
               if ((strucmp(p, archfile)) != 0)         /* not self */
                  found += stash_name(p);
         }
   if (zflag)
      while (gets(buf))
            if (buf[0] != '\0')
            found += stash_name(buf);
   return (found);
   }


/*page*/
/*
** squirrel a name away in the linked list of targets
*/

stash_name(p)
char  *p;
   {
   static   FN    *fnp;
   char           *q;

   if (*p == '/')
      fatal(1, "absolute path illegal <%s>\n", p);
   q = emalloc(sizeof(FN) + strlen(p));
   if (fnhead == (FN *)NULL)
      fnhead = fnp = (FN *)q;
   else
      fnp = fnp->fn_link = (FN *)q;
   strcpy(fnp->fn_name, p);
   fnp->fn_link = (FN *)NULL;
   return (1);                                 /* we saved one name */
   }
/*page*/
/*
** get the next header from the file, perhaps skipping the body
*/

get_header(fp, hp)
FILE     *fp;
HEADER   *hp;
   {
   long count;

   if ((count = (long) fread(hp, 1, sizeof(HEADER), fp)) == NULL)
      return (EOF);
   if (strncmp(hp->a_hid, hid, HIDSIZ) != 0)
    if (hp->a_hid[0] && hp->a_hid[0]!=26)   /* check for Xmodem padding */
      fatal(1, "file not archive or damaged\n");
    else
    {
      fseek(fp, - (count), 1);   /* backup over Xmodem padding */
      return (EOF);                     /* and return EOF */
    }
   return (0);
   }


/*
** here we will recreate a tree that was collapsed into the archive file
*/

FILE  *spl_open(hp)
HEADER   *hp;
   {
   char  buf[FNSIZ+3];
   FILE  *ofp;
   char  *p;
   long  c4tol();

   p = hp->a_name;
   while (p = index(p, '/'))
      {
      *p = '\0';                            /* truncate temporarily */
      if (access(hp->a_name, DIRCHK) == -1)
         if (mknod(hp->a_name, DIRMAK) == -1)     /* try to make it */
            fatal(errno, "can't make <%s>\n", hp->a_name);
      *p++ = '/';                             /* put back the delim */
      }

   strcpy(buf, hp->a_name);
   if (hp->a_stat)
      sprintf(&buf[strlen(buf)], ".%d", hp->a_stat); /* make  unique */
   if ((ofp = fopen(buf, "w")) == NULL)
      fatal(errno, "create failure on %s\n", buf);
   set_fsize(ofp->_fd, c4tol(hp->a_attr.fd_fsize));
   return (ofp);
   }
/*page*/
/*
** copy an archived file from an archive
*/

copy_from(ifp, ofp, hp)
FILE     *ifp, *ofp;
HEADER   *hp;
   {
   long  bytes = hp->a_size;
   int   byt;

   switch (hp->a_type)
      {
      case PLAIN :
         while (bytes--)
            {
            if ((byt = getc(ifp)) == ERROR)
               fatal(errno, "read error while extracting\n");
            if (putc(byt, ofp) == ERROR)
               fatal(errno, "write error while extracting\n");
            }
         break;

      case COMP1 :
         if ((byt=de_LZ_1(ifp, ofp, bytes,11)) == -2)
            fatal(1, "archive damaged or file not compressed\n");
         break;

      case COMP2 :
         if ((byt=de_LZ_1(ifp, ofp, bytes,12)) == -2)
            fatal(1, "archive damaged or file not compressed\n");
         break;

      default :
         fatal(1, "unknown compression mode\n");
      }
   }
/*page*/
/*
** copy an file to an archive
*/

long copy_to(ofp, ifp, hp)
FILE     *ofp, *ifp;
HEADER   *hp;
   {
   long  bytes = 0;
   int   byt;

   switch (hp->a_type)
      {
      case PLAIN :
         while ((byt = getc(ifp)) != ERROR)
            if (putc(byt, ofp) == ERROR)
               fatal(errno, "write error while archiving\n");
            else
               ++bytes;
         if (ferror(ifp))
            fatal(errno, "read error while archiving\n");
         break;

      case COMP1 :
         if ((byt = LZ_1(ifp, ofp, &bytes,11)) == -3)
            fatal(1, "string table overflow on compression\n");
         break;

      case COMP2 :
         if ((byt = LZ_1(ifp, ofp, &bytes,12)) == -3)
            fatal(1, "string table overflow on compression\n");
         break;

      default :
         fatal(1, "unknown compression mode\n");
      }
   return (bytes);
   }
/*page*/
/*
** get memory from the system or die trying
*/

char  *emalloc(n)
int   n;
   {
   char  *p;
   char *malloc();
   
   if ((p = malloc(n)) == NULL)
      fatal(errno, "Can't allocate memory\n");
   return (p);
   }


/*
** help for the user
*/

help()
   {
   char *sp=" ";

   fprintf(stderr, "%s", title);
   fprintf(stderr, "Usage:  Ar -<cmd>[<modifier>] [file .. ]\n");
   fprintf(stderr, "      <cmd> is one of the following:\n");
   fprintf(stderr, "%9s%s", sp, "t  show table of contents for archive\n");
   fprintf(stderr, "%9s%s", sp, "u  update/add file(s) to the archive\n");
   fprintf(stderr, "%9s%s", sp, "p  print file(s) from the archive\n");
   fprintf(stderr, "%9s%s", sp, "x  extract file(s) from the archive\n");
   fprintf(stderr, "      <modifier> is one of the following:\n");
   fprintf(stderr, "%9s%s", sp, "a  all versions (for extract)\n");
   fprintf(stderr, "%9s%s", sp, "o  oldstyle compression (11 bit)\n");
   fprintf(stderr, "%9s%s", sp, "s  supress file compression\n");
   fprintf(stderr, "%9s%s", sp, "z  read names for <cmd> from std in\n");
   fprintf(stderr, "\n      File names can include the meta chars ");
   fprintf(stderr, "* and ?, or path lists.\n");
   exit (1);
   }


/*
** print a fatal error message and exit
*/

fatal(code, msg, arg1, arg2)
int   code;
char  *msg;
int   arg1, arg2;
   {
   fprintf(stderr, "%s: ", mod);
   fprintf(stderr, msg, arg1, arg2);
   exit (code);
   }

show_head()
{
   printf("%55s", title);
   barline();
}

main(argc, argv)
int   argc;
char  **argv;
{
   char  command, *p;
   int   n, updating;
   FILE  *afp;
   char *strcat();

   pflinit();

   mod = *argv++;                              /* save program name */

   if ((argc < 3) || (*(p = *argv++) != '-'))
      help();

   command = tolower(*++p);

   proc_opt(++p);                      /* process command modifiers */

   archfile = *argv++;
   n = strlen(archfile);
   if ((strcmp(archfile + n - SUFSIZ, suf)) != 0)
      archfile = strcat(strcpy(emalloc(n + SUFSIZ + 1), archfile), suf);

   if (get_names(argc -= 3, argv, command == 'u') == 0)
      if (command == 'u')
         fatal(0, "none of the targets found\n");

   if (command == 'u')
      {
      if ((afp = fopen(archfile, "r+")) == NULL)   /* try old first */
         afp = fopen(archfile, "w+");      /* not there - create it */
      }
   else
      {
      if ((afp = fopen(archfile, "r")) == NULL)   /* save file date */
         fatal(errno, "can't find %s\n", archfile);
      }

   proc_cmd(command, afp);                     /* process a command */

   return 0;
}
